from __future__ import annotations
from typing import Dict

from pacco.manager.interfaces.package_manager import PackageManagerInterface


class RemoteInterface:
    def __init__(self, name: str, remote_type: str, package_manager: PackageManagerInterface):
        self.name = name
        self.remote_type = remote_type
        self.package_manager = package_manager

    def __str__(self):
        return "[{}, {}]".format(self.name, self.remote_type)

    @staticmethod
    def create(name: str, serialized: Dict[str, str]) -> RemoteInterface:
        raise NotImplementedError()

    def serialize(self) -> Dict[str, str]:
        raise NotImplementedError()



import logging
import os
import re
import shutil
from pathlib import Path
from typing import Dict

from pacco.cli.commands.utils.command_abstract import CommandAbstract
from pacco.manager.utils.cache import Cache


class Binary(CommandAbstract):
    @staticmethod
    def __parse_settings_args(settings_args: str) -> Dict[str, str]:
        if not re.match(r"([\w\-.]+=[\w\-.]+,)*([\w\-.]+=[\w\-.]+),?", settings_args):
            raise ValueError("The settings configuration must match ([\\w-.]+=[\\w-.]+,)*([\\w-.]+=[\\w-.]+),?")
        return {token.split('=')[0]: token.split('=')[1] for token in settings_args.split(',')}

    def download(self, *args):
        parser = self.init_parser('download')
        parser.add_argument("remote_name", help="remote name")
        parser.add_argument("registry_name", help="registry name")
        parser.add_argument("dir_path", help="download path")
        parser.add_argument("settings", help="settings for the specified registry "
                                             "(e.g. os=linux,version=2.1.0,type=debug")
        parsed_args = parser.parse_args(args)

        settings_dict = Binary.__parse_settings_args(parsed_args.settings)
        if parsed_args.remote_name == 'default':
            self.rm.default_download(parsed_args.registry_name, settings_dict, parsed_args.dir_path)
        pm = self.rm.get_remote(parsed_args.remote_name)
        pr = pm.get_package_registry(parsed_args.registry_name)
        pb = pr.get_package_binary(settings_dict)
        pb.download_content(parsed_args.dir_path)

    def upload(self, *args):
        parser = self.init_parser('upload')
        parser.add_argument("remote_name", help="remote name")
        parser.add_argument("registry_name", help="registry name")
        parser.add_argument("dir_path", help="directory to be uploaded")
        parser.add_argument("settings", help="settings for the specified registry "
                                             "(e.g. os=linux,version=2.1.0,type=debug")
        parsed_args = parser.parse_args(args)

        assignment = Binary.__parse_settings_args(parsed_args.settings)
        pm = self.rm.get_remote(parsed_args.remote_name)
        pr = pm.get_package_registry(parsed_args.registry_name)
        try:
            pr.get_package_binary(assignment)
        except FileNotFoundError:
            pr.add_package_binary(assignment)
        else:
            self.out.writeln("WARNING: Existing binary found, overwriting")
        finally:
            pb = pr.get_package_binary(assignment)
            pb.upload_content(parsed_args.dir_path)

    def remove(self, *args):
        parser = self.init_parser('remove')
        parser.add_argument("remote_name", help="remote name")
        parser.add_argument("registry_name", help="registry name")
        parser.add_argument("settings", help="settings for the specified registry "
                                             "(e.g. os=linux,version=2.1.0,type=debug")
        parsed_args = parser.parse_args(args)

        assignment = Binary.__parse_settings_args(parsed_args.settings)
        pm = self.rm.get_remote(parsed_args.remote_name)
        pr = pm.get_package_registry(parsed_args.registry_name)
        pr.remove_package_binary(assignment)

    def reassign(self, *args):
        """
        Change the assignment of a binary to a new one
        """
        parser = self.init_parser('reassign')
        parser.add_argument("remote_name", help="remote name")
        parser.add_argument("registry_name", help="registry name")
        parser.add_argument("old_settings", help="old settings (e.g. os=linux,version=2.1.0,type=debug")
        parser.add_argument("new_settings", help="new settings (e.g. os=osx,version=2.1.1,type=debug")
        parsed_args = parser.parse_args(args)
        old_assignment = Binary.__parse_settings_args(parsed_args.old_settings)
        new_assignment = Binary.__parse_settings_args(parsed_args.new_settings)
        pm = self.rm.get_remote(parsed_args.remote_name)
        pr = pm.get_package_registry(parsed_args.registry_name)
        pr.reassign_binary(old_assignment, new_assignment)

    def get_location(self, *args):
        parser = self.init_parser('get_location')
        parser.add_argument("registry_name", help="registry name")
        parser.add_argument("settings", help="settings for the specified registry "
                                             "(e.g. os=linux,version=2.1.0,type=debug")
        parser.add_argument("--fresh_download", help="add this flag to not use local cache",
                            action="store_true")
        parsed_args = parser.parse_args(args)
        assignment = Binary.__parse_settings_args(parsed_args.settings)
        if not parsed_args.fresh_download:
            try:
                self.out.write(Cache().get_path(parsed_args.registry_name, assignment))
                return
            except ValueError:
                logging.warning("The binary is not found in cache, will attemp fresh download")
        download_path = os.path.join(str(Path.home()), '.pacco_tmp')
        self.rm.default_download(parsed_args.registry_name, assignment,
                                 download_path, fresh_download=parsed_args.fresh_download)
        shutil.rmtree(download_path)
        self.out.write(Cache().get_path(parsed_args.registry_name, assignment))



from typing import List

from pacco.manager.interfaces.package_registry import PackageRegistryInterface


class PackageManagerInterface:
    """
    Represent the existence of the manager in a remote. This class is the interface class with the
    expected behavior defined below.
    """

    def __init__(self):
        pass

    def list_package_registries(self) -> List[str]:
        """
        List package registries in this package manager.

        Returns:
            The list of package registry name
        """
        raise NotImplementedError()

    def remove_package_registry(self, name: str) -> None:
        """
        Delete a package registry from the package manager.

        Args:
            name: the name of the package registry to be deleted.
        """
        raise NotImplementedError()

    def add_package_registry(self, name: str, params: List[str]) -> None:
        """
        Add a new package registry to this package manager.

        Args:
            name: the name of the package. For printing purposes only.
            params: the list of keys for the configuration parameter, e.g. ['os', 'compiler', 'version']
        Exception:
            FileExistsError: raised if the package with the same name is found
        """
        raise NotImplementedError()

    def get_package_registry(self, name: str) -> PackageRegistryInterface:
        """
        Get a reference to the ``PackageRegistry`` object based on the settings value

        Args:
            name: the name of the package registry to get
        Returns:
            the object
        Exceptions:
            FileNotFoundError: when that package is not found or it is not set properly.
        """
        raise NotImplementedError()


import logging

from typing import Optional, Dict

from pacco.manager.utils.cache import Cache
from pacco.manager.utils.clients import FileBasedClientAbstract
from pacco.manager.interfaces.package_binary import PackageBinaryInterface


class PackageBinaryFileBased(PackageBinaryInterface):
    """
    An implementation of the PackageBinary interface

    Examples:
        >>> from pacco.manager.utils.clients import LocalClient, NexusFileClient
        >>> client = LocalClient(clean=True)
        >>> import os
        >>> if 'NEXUS_URL' in os.environ: client = NexusFileClient(os.environ['NEXUS_URL'], 'admin', 'admin123', clean=True)
        >>> from pacco.manager.file_based.package_manager import PackageManagerFileBased
        >>> pm = PackageManagerFileBased(client)
        >>> pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        >>> pr = pm.get_package_registry('openssl')
        >>> pr.add_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        >>> pb = pr.get_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        >>> import os, shutil
        >>> os.makedirs('testfolder', exist_ok=True)
        >>> open('testfolder/testfile', 'w').close()
        >>> pb.upload_content('testfolder')
        >>> __ = shutil.move('testfolder/testfile', 'testfolder/testfile2')
        >>> pb_get = pr.get_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})  # use a new reference
        >>> pb_get.download_content('testfolder')
        >>> sorted(os.listdir('testfolder'))
        ['testfile', 'testfile2']
        >>> shutil.rmtree('testfolder')
    """

    def __init__(self, client: FileBasedClientAbstract, registry_name: Optional[str] = None,
                 assignment: Optional[Dict[str, str]] = None):
        if not isinstance(client, FileBasedClientAbstract):
            raise TypeError("Must be using FileBasedClient")
        self.client = client
        super(PackageBinaryFileBased, self).__init__()
        self.__registry_name = registry_name
        self.__cache = Cache()
        self.__assignment = assignment
        self.__cache_enabled = bool(self.__registry_name) and bool(self.__assignment)

    def __repr__(self):
        return "PackageBinaryObject"

    def download_content(self, download_dir_path: str, fresh_download: Optional[bool] = False) -> None:
        if self.__cache_enabled and (not fresh_download) and \
                self.__cache.download_from_cache(self.__registry_name, self.__assignment, download_dir_path):
            logging.info("use cache")
            return
        logging.info("fresh download")
        self.client.download_dir(download_dir_path)
        if self.__cache_enabled:
            logging.info("save cache")
            self.__cache.upload_to_cache(self.__registry_name, self.__assignment, download_dir_path)

    def upload_content(self, dir_path: str) -> None:
        self.client.upload_dir(dir_path)
        if self.__cache_enabled:
            self.__cache.upload_to_cache(self.__registry_name, self.__assignment, dir_path)


from pacco.cli.commands.utils.command_abstract import CommandAbstract
from pacco.cli.commands import remote, registry, binary
from pacco.cli.commands.utils.output_stream import OutputStream
from pacco.manager.remote_manager import RemoteManager


class Pacco(CommandAbstract):
    def remote(self, *args: str):
        remote.Remote('remote', self.out, self.rm).run(*args)

    def registry(self, *args: str):
        registry.Registry('registry', self.out, self.rm).run(*args)

    def binary(self, *args: str):
        binary.Binary('binary', self.out, self.rm).run(*args)


def main(args):
    Pacco('', OutputStream(), RemoteManager()).run(*args)



from pacco.cli.commands.utils.command_abstract import CommandAbstract
from pacco.manager.remote_manager import ALLOWED_REMOTE_TYPES


class Remote(CommandAbstract):
    def list(self, *args):
        """
        List existing remotes.
        """
        parser = self.init_parser('list')
        parser.parse_args(args)
        remotes = self.rm.list_remote()
        self.out.writeln(remotes)

    def add(self, *args):
        """
        Add a remote.
        """
        parser = self.init_parser('add')
        parser.add_argument("name", help="remote name")
        parser.add_argument("type", help="remote type", choices=ALLOWED_REMOTE_TYPES)
        parsed_args = parser.parse_args(args)
        if parsed_args.type == "local":
            path = input("Path (if empty, ~/.pacco/ will be used): ")
            self.rm.add_remote(parsed_args.name, {
                "remote_type": "local",
                "path": path
            })
        elif parsed_args.type == "nexus_site":
            url = input("URL: ")
            username = input("Username: ")
            from getpass import getpass
            password = getpass()
            self.rm.add_remote(parsed_args.name, {
                "remote_type": "nexus_site",
                "url": url,
                "username": username,
                "password": password
            })

    def remove(self, *args):
        """
        Remove a remote.
        """
        parser = self.init_parser('remove')
        parser.add_argument("name", help="remote name")
        parsed_args = parser.parse_args(args)
        self.rm.remove_remote(parsed_args.name)

    def set_default(self, *args):
        """
        Set default remote(s).
        """
        parser = self.init_parser('set_default')
        parser.add_argument("name", nargs="*", help="remote name")
        parsed_args = parser.parse_args(args)
        self.rm.set_default(parsed_args.name)

    def list_default(self, *args):
        """
        List default remote(s).
        """
        parser = self.init_parser('list_default')
        parser.parse_args(args)
        default_remotes = self.rm.get_default()
        self.out.writeln(default_remotes)



import os
from pathlib import Path

from typing import Dict

from pacco.manager.utils.clients import LocalClient


class Cache:
    def __init__(self, clean=False):
        self.client = LocalClient(os.path.join(str(Path.home()), '.pacco_cache'), clean=clean)

    @staticmethod
    def __serialize(registry_name: str, assignment: Dict[str, str]):
        assignment['__pacco_registry_name'] = registry_name
        sorted_assignment_tuple = sorted(assignment.items(), key=lambda x: x[0])
        zipped_assignment = ['='.join(pair) for pair in sorted_assignment_tuple]
        return '=='.join(zipped_assignment)

    def download_from_cache(self, registry_name: str, assignment: Dict[str, str], download_path: str) -> bool:
        dir_name = Cache.__serialize(registry_name, assignment)
        if dir_name in self.client.ls():
            self.client.dispatch_subdir(dir_name).download_dir(download_path)
            return True
        else:
            return False

    def upload_to_cache(self, registry_name: str, assignment: Dict[str, str], source_path: str) -> None:
        dir_name = Cache.__serialize(registry_name, assignment)
        os.makedirs(dir_name, exist_ok=True)
        self.client.dispatch_subdir(dir_name).upload_dir(source_path)

    def get_path(self, registry_name: str, assignment: Dict[str, str]) -> str:
        dir_name = Cache.__serialize(registry_name, assignment)
        if dir_name in self.client.ls():
            return self.client.dispatch_subdir(dir_name).bin_dir_for_cache
        raise ValueError("No cache found")



import re

from pacco.cli.commands.utils.command_abstract import CommandAbstract


class Registry(CommandAbstract):
    def list(self, *args):
        """
        List registries of a remote.
        """
        parser = self.init_parser('list')
        parser.add_argument("remote", help="remote name")
        parsed_args = parser.parse_args(args)
        pm = self.rm.get_remote(parsed_args.remote)
        self.out.writeln(pm.list_package_registries())

    def add(self, *args):
        """
        Add registry to remote.
        """
        parser = self.init_parser('add')
        parser.add_argument("remote", help="remote name")
        parser.add_argument("name", help="registry name")
        parser.add_argument("settings", help="settings key (e.g. os,version,obfuscation)")
        parsed_args = parser.parse_args(args)
        if not re.match(r"([(\w)-.]+,)*([(\w)-.]+),?", parsed_args.settings):
            raise ValueError("Settings must be in the form of ([(\\w)-.]+,)*([(\\w)-.]+),?")
        pm = self.rm.get_remote(parsed_args.remote)
        pm.add_package_registry(parsed_args.name, parsed_args.settings.split(","))

    def remove(self, *args):
        """
        Remove a registry from a specific remote.
        """
        parser = self.init_parser('remove')
        parser.add_argument("remote", help="remote name")
        parser.add_argument("name", help="registry name")
        parsed_args = parser.parse_args(args)
        pm = self.rm.get_remote(parsed_args.remote)
        pm.remove_package_registry(parsed_args.name)

    def binaries(self, *args):
        """
        List binaries of a registry from a specific remote.
        """
        parser = self.init_parser('binaries')
        parser.add_argument("remote", help="remote name")
        parser.add_argument("name", help="registry name")
        parsed_args = parser.parse_args(args)
        pm = self.rm.get_remote(parsed_args.remote)
        pr = pm.get_package_registry(parsed_args.name)
        self.out.writeln(pr.list_package_binaries())

    def param_list(self, *args):
        """
        List params of a registry.
        """
        parser = self.init_parser('param_list')
        parser.add_argument("remote", help="remote name")
        parser.add_argument("name", help="registry name")

        parsed_args = parser.parse_args(args)
        pm = self.rm.get_remote(parsed_args.remote)
        pr = pm.get_package_registry(parsed_args.name)
        self.out.writeln(pr.param_list())

    def param_add(self, *args):
        """
        Add new parameter with default value to the binaries.
        """
        parser = self.init_parser('param_add')
        parser.add_argument("remote", help="remote name")
        parser.add_argument("name", help="registry name")
        parser.add_argument("param_name", help="the new param name to be added")
        parser.add_argument("default_value", help="the default_value assigned to the new param")

        parsed_args = parser.parse_args(args)
        pm = self.rm.get_remote(parsed_args.remote)
        pr = pm.get_package_registry(parsed_args.name)
        pr.param_add(parsed_args.param_name, parsed_args.default_value)

    def param_remove(self, *args):
        """
        Remove an existing parameter from all binaries.
        """
        parser = self.init_parser('param_remove')
        parser.add_argument("remote", help="remote name")
        parser.add_argument("name", help="registry name")
        parser.add_argument("param_name", help="the param name to be removed")

        parsed_args = parser.parse_args(args)
        pm = self.rm.get_remote(parsed_args.remote)
        pr = pm.get_package_registry(parsed_args.name)
        pr.param_remove(parsed_args.param_name)



from typing import Optional, List, Dict

from pacco.manager.interfaces.package_binary import PackageBinaryInterface


class PackageRegistryInterface:
    """
    Represent the existence of a package (e.g. openssl) in the package manager.
    This class is the interface class with the expected behavior defined below.
    """

    def __init__(self, name: str, params: Optional[List[str]] = None):
        self.name = name
        self.params = params

    def list_package_binaries(self) -> List[Dict[str, str]]:
        """
        List the package binaries registered in this package registry

        Returns:
            list of the package binary assignment dictionaries
        """
        raise NotImplementedError()

    def add_package_binary(self, assignment: Dict[str, str]) -> None:
        """
        Add a new package binary to this registry. Note that this will only declare the existence of the binary
        by creating a new directory, to upload the binary must be done through the ``PackageBinaryFileBased``
        object itself.

        Args:
            assignment: the assignment of key value of the params.
        Exceptions:
            KeyError: raised if the set of keys in the passed ``assignment`` is different with ``params``
            FileExistsError: raised if a package binary with the same configuration already exist.
        """
        raise NotImplementedError()

    def remove_package_binary(self, assignment: Dict[str, str]):
        """
        Delete the package binary folder

        Args:
            assignment: the configuration of the the package binary to be deleted
        """
        raise NotImplementedError()

    def get_package_binary(self, assignment: Dict[str, str]) -> PackageBinaryInterface:
        """
        Get a reference to the ``PackageBinary`` object based on the settings value

        Args:
            assignment: the configuration of the the package binary to get
        Returns:
            the object
        Exceptions:
            KeyError: when the key of the settings passed is not correct
            FileNotFoundError: when there is no binary with the configuration of settings value
        """
        raise NotImplementedError()

    def param_list(self) -> List[str]:
        """
        List the declared parameters of the ```PackageRegistry```
        """
        raise NotImplementedError()

    def param_add(self, name: str, default_value: Optional[str] = "default") -> None:
        """
        Append new parameter to each ``PackageBinary`` object and assign ``default_value`` as
        the default value to the new parameter

        Args:
            name: the new param name
            default_value: the default value to be assigned
        Exceptions:
            ValueError: if the param is already exist
        """
        raise NotImplementedError()

    def param_remove(self, name: str) -> None:
        """
        Remove a parameter from each ``PackageBinary`` object

        Args:
            name: the param name to be deleted
        Exceptions:
            ValueError: if the param name does not exist
            NameError: if the resulting assignments will have duplicate when the param is removed
        """
        raise NotImplementedError()

    def reassign_binary(self, old_assignment: Dict[str, str], new_assignment: Dict[str, str]) -> None:
        """
        Reassign a new assignment to an existing binary

        Args:
            old_assignment: the old assignment
            new_assignment: the new assignment
        Exceptions:
            KeyError: if the key in the new assignment does not match with params
            ValueError: if there is no binary that match old_assignment
            NameError: there already exist binary with the same configuration as new_assignment
        """
        raise NotImplementedError()



from __future__ import annotations
import os
from typing import Dict, Optional

from pacco.manager.file_based.package_manager import PackageManagerFileBased
from pacco.manager.interfaces.remote import RemoteInterface
from pacco.manager.utils.clients import LocalClient, NexusFileClient, FileBasedClientAbstract


class RemoteFileBased(RemoteInterface):
    def __init__(self, name: str, remote_type: str, client: FileBasedClientAbstract):
        self.client = client
        super(RemoteFileBased, self).__init__(name, remote_type, PackageManagerFileBased(client))

    @staticmethod
    def create(name: str, serialized: Dict[str, str]) -> RemoteInterface:
        raise NotImplementedError()

    def serialize(self) -> Dict[str, str]:
        raise NotImplementedError()


class LocalRemote(RemoteFileBased):
    def __init__(self, name: str, remote_type: str, path: Optional[str] = "", clean: Optional[bool] = False):
        if path:
            self.__path = os.path.abspath(path)
        else:
            self.__path = ""
        client = LocalClient(self.__path, clean)
        super(LocalRemote, self).__init__(name, remote_type, client)

    @staticmethod
    def create(name: str, serialized: Dict[str, str]) -> LocalRemote:
        if 'path' not in serialized:
            serialized['path'] = ""
        return LocalRemote(name, serialized['remote_type'], serialized['path'])

    def serialize(self) -> Dict[str, str]:
        return {'remote_type': 'local', 'path': self.__path}


class NexusSiteRemote(RemoteFileBased):
    url: str
    username: str
    password: str

    def __init__(self, name: str, remote_type: str, client: NexusFileClient):
        super(NexusSiteRemote, self).__init__(name, remote_type, client)

    @staticmethod
    def create(name: str, serialized: Dict[str, str]) -> NexusSiteRemote:
        client = NexusFileClient(serialized['url'], serialized['username'], serialized['password'])
        remote_object = NexusSiteRemote(name, serialized['remote_type'], client)
        remote_object.url = serialized['url']
        remote_object.username = serialized['username']
        remote_object.password = serialized['password']
        return remote_object

    def serialize(self) -> Dict[str, str]:
        return {'remote_type': 'nexus_site', 'url': self.url,
                'username': self.username, 'password': self.password}



from typing import List

from pacco.manager.utils.clients import FileBasedClientAbstract
from pacco.manager.file_based.package_registry import PackageRegistryFileBased
from pacco.manager.interfaces.package_manager import PackageManagerInterface


class PackageManagerFileBased(PackageManagerInterface):
    """
    An implementation of the PackageManager interface

    Examples:
        >>> from pacco.manager.utils.clients import LocalClient, NexusFileClient
        >>> client = LocalClient(clean=True)
        >>> import os
        >>> if 'NEXUS_URL' in os.environ: client = NexusFileClient(os.environ['NEXUS_URL'], 'admin', 'admin123', clean=True)
        >>> pm = PackageManagerFileBased(client)
        >>> pm.list_package_registries()
        []
        >>> pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        >>> pm.add_package_registry('boost', ['os', 'target', 'type'])
        >>> pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        Traceback (most recent call last):
            ...
        FileExistsError: The package registry openssl is already found
        >>> pm.list_package_registries()
        ['boost', 'openssl']
        >>> pm.remove_package_registry('openssl')
        >>> pm.list_package_registries()
        ['boost']
        >>> pm.get_package_registry('boost')
        PR[boost, os, target, type]
    """

    def __init__(self, client: FileBasedClientAbstract):
        if not isinstance(client, FileBasedClientAbstract):
            raise TypeError("Must be using FileBasedClient")
        self.client = client
        super(PackageManagerFileBased, self).__init__()

    def list_package_registries(self) -> List[str]:
        return sorted(self.client.ls())

    def remove_package_registry(self, name: str) -> None:
        self.client.rmdir(name)

    def add_package_registry(self, name: str, params: List[str]) -> None:
        dirs = self.client.ls()
        if name in dirs:
            raise FileExistsError("The package registry {} is already found".format(name))
        self.client.mkdir(name)
        PackageRegistryFileBased(name, self.client.dispatch_subdir(name), params)
        return

    def get_package_registry(self, name: str) -> PackageRegistryFileBased:
        dirs = self.client.ls()
        if name not in dirs:
            raise FileNotFoundError("The package registry {} is not found".format(name))
        return PackageRegistryFileBased(name, self.client.dispatch_subdir(name))

    def __repr__(self):
        return "PackageManagerObject"



from typing import Optional


class PackageBinaryInterface:
    """
        Represent the existence of a package (e.g. openssl) in the package manager
        This class is the interface class with the expected behavior defined below.
    """

    def __init__(self):
        pass

    def download_content(self, download_dir_path: str, fresh_download: Optional[bool] = False) -> None:
        """
        Download content of uploaded binary from the remote to the ``download_dir_path``

        Args:
            download_dir_path: the destination of download
            fresh_download: if true, will not use cache
        """
        raise NotImplementedError()

    def upload_content(self, dir_path: str) -> None:
        """
        Remove the previous binary and upload the content of ``dir_path`` to the remote.

        Args:
            dir_path: the path to the directory to be uploaded
        """
        raise NotImplementedError()



from pacco.manager.file_based.remote import LocalRemote, NexusSiteRemote


def instantiate_remote(name: str, serialized):
    if serialized['remote_type'] == 'local':
        return LocalRemote.create(name, serialized)
    elif serialized['remote_type'] == 'nexus_site':
        return NexusSiteRemote.create(name, serialized)
    else:
        raise ValueError("The remote_type {} is not supported, currently only supports [{}]".format(
            serialized['remote_type'], ", ".join(['local', 'nexus_site'])
        ))



import random
import re
import string

from typing import Optional, List, Dict, Callable

from pacco.manager.utils.clients import FileBasedClientAbstract
from pacco.manager.file_based.package_binary import PackageBinaryFileBased
from pacco.manager.interfaces.package_registry import PackageRegistryInterface


class PackageRegistryFileBased(PackageRegistryInterface):
    """
    An implementation of the PackageRegistry interface

    Examples:
        >>> from pacco.manager.utils.clients import LocalClient, NexusFileClient
        >>> client = LocalClient(clean=True)
        >>> import os
        >>> if 'NEXUS_URL' in os.environ: client = NexusFileClient(os.environ['NEXUS_URL'], 'admin', 'admin123', clean=True)
        >>> from pacco.manager.file_based.package_manager import PackageManagerFileBased
        >>> pm = PackageManagerFileBased(client)
        >>> pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        >>> pr = pm.get_package_registry('openssl')
        >>> pr.list_package_binaries()
        []
        >>> pr.add_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        >>> pr.add_package_binary({'host_os':'osx', 'compiler':'clang', 'version':'1.0'})
        Traceback (most recent call last):
            ...
        KeyError: "wrong settings key: ['compiler', 'host_os', 'version'] is not ['compiler', 'os', 'version']"
        >>> pr.add_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        Traceback (most recent call last):
            ...
        FileExistsError: such binary already exist
        >>> len(pr.list_package_binaries())
        1
        >>> pr.add_package_binary({'os':'linux', 'compiler':'gcc', 'version':'1.0'})
        >>> len(pr.list_package_binaries())
        2
        >>> pr.remove_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        >>> len(pr.list_package_binaries())
        1
        >>> pr.get_package_binary({'os':'linux', 'compiler':'gcc', 'version':'1.0'})
        PackageBinaryObject
        >>> pr.add_package_binary({'os':'linux', 'compiler':'g++', 'version':'1.0'})
        >>> pr.param_remove('compiler')
        Traceback (most recent call last):
        ...
        NameError: Cannot remove parameter compiler since it will cause two binary to have the same value
        >>> pr.param_add('stdlib', default_value='c++11')
        >>> pr
        PR[openssl, compiler, os, stdlib, version]
        >>> old_assignment = {'os':'linux', 'compiler':'g++', 'version':'1.0', 'stdlib': 'c++11'}
        >>> new_assignment = {'os':'linux', 'compiler':'g++', 'version':'1.0', 'stdlib': 'static_c++'}
        >>> pr.reassign_binary(old_assignment, new_assignment)
        >>> pr.param_remove('compiler')
        >>> pr
        PR[openssl, os, stdlib, version]
    """
    __params_prefix = '__params'

    def __init__(self, name: str, client: FileBasedClientAbstract, params: Optional[List[str]] = None):
        if not isinstance(client, FileBasedClientAbstract):
            raise TypeError("Must be using FileBasedClient")
        self.client = client
        super(PackageRegistryFileBased, self).__init__(name, params)

        remote_params = self.__get_remote_params()
        if params is None and remote_params is None:
            raise FileNotFoundError("you need to declare params if you are adding. if you are getting, this "
                                    "means that the package registry is not properly set, you need to delete and "
                                    "add again")
        elif remote_params is not None:  # ignore the passed params and use the remote one
            self.params = remote_params
        else:
            self.params = params
            self.client.mkdir(self.__serialize_params(self.params))

    def __repr__(self):
        return "PR[{}, {}]".format(self.name, ', '.join(sorted(self.params)))

    def __get_remote_params(self) -> Optional[List[str]]:
        params = None
        dirs = self.client.ls()
        for dir_name in dirs:
            if PackageRegistryFileBased.__params_prefix in dir_name:
                params = dir_name.split('==')[1:]
        return params

    @staticmethod
    def __serialize_params(params: List[str]) -> str:
        params = sorted(params)
        return '=='.join([PackageRegistryFileBased.__params_prefix] + params)

    @staticmethod
    def __serialize_assignment(assignment: Dict[str, str]) -> str:
        for key, value in assignment.items():
            if len(value) == 0:
                raise ValueError("assignment value for param {} cannot be an empty string".format(key))
        sorted_assignment_tuple = sorted(assignment.items(), key=lambda x: x[0])
        zipped_assignment = ['='.join(pair) for pair in sorted_assignment_tuple]
        return '=='.join(zipped_assignment)

    @staticmethod
    def __unserialize_assignment(dir_name: str) -> Dict[str, str]:
        if not re.match(r"((\w+=\w+)==)*(\w+=\w+)", dir_name):
            raise ValueError("Invalid dir_name syntax {}".format(dir_name))
        return {arg.split('=')[0]: arg.split('=')[1] for arg in dir_name.split('==')}

    def __get_serialized_assignment_to_wrapper_mapping(self):
        dir_names = self.client.ls()
        dir_names.remove(self.__serialize_params(self.params))

        mapping = {}
        for dir_name in dir_names:
            sub_dirs = self.client.dispatch_subdir(dir_name).ls()
            if 'bin' in sub_dirs:
                sub_dirs.remove('bin')
            serialized_assignment = sub_dirs[0]
            mapping[serialized_assignment] = dir_name

        return mapping

    def list_package_binaries(self) -> List[Dict[str, str]]:
        return [PackageRegistryFileBased.__unserialize_assignment(serialized_assignment)
                for serialized_assignment in self.__get_serialized_assignment_to_wrapper_mapping()]

    @staticmethod
    def __random_string(length: int) -> str:
        return ''.join(random.choices(string.ascii_uppercase + string.digits, k=length))

    def add_package_binary(self, assignment: Dict[str, str]) -> None:
        if set(assignment.keys()) != set(self.params):
            raise KeyError("wrong settings key: {} is not {}".format(sorted(assignment.keys()),
                                                                     sorted(self.params)))

        serialized_assignment = PackageRegistryFileBased.__serialize_assignment(assignment)
        mapping = self.__get_serialized_assignment_to_wrapper_mapping()
        if serialized_assignment in mapping:
            raise FileExistsError("such binary already exist")

        new_random_dir_name = PackageRegistryFileBased.__random_string(10)
        if new_random_dir_name in mapping.values():
            new_random_dir_name = PackageRegistryFileBased.__random_string(10)

        self.client.mkdir(new_random_dir_name)
        self.client.dispatch_subdir(new_random_dir_name).mkdir(serialized_assignment)
        return

    def remove_package_binary(self, assignment: Dict[str, str]):
        self.client.rmdir(self.__get_serialized_assignment_to_wrapper_mapping()[
                              PackageRegistryFileBased.__serialize_assignment(assignment)
                          ])

    def get_package_binary(self, assignment: Dict[str, str]) -> PackageBinaryFileBased:
        serialized_assignment = PackageRegistryFileBased.__serialize_assignment(assignment)
        if set(assignment.keys()) != set(self.params):
            raise KeyError("wrong settings key: {} is not {}".format(sorted(assignment.keys()),
                                                                     sorted(self.params)))
        if serialized_assignment not in self.__get_serialized_assignment_to_wrapper_mapping():
            raise FileNotFoundError("such configuration does not exist")
        return PackageBinaryFileBased(
            self.client.dispatch_subdir(
                self.__get_serialized_assignment_to_wrapper_mapping()[serialized_assignment]
            ),
            registry_name=self.name,
            assignment=assignment,
        )

    def __rename_serialized_assignment(self, action: Callable[[Dict[str, str]], None]):
        for serialized_assignment, dir_name in self.__get_serialized_assignment_to_wrapper_mapping().items():
            assignment = PackageRegistryFileBased.__unserialize_assignment(serialized_assignment)
            action(assignment)
            new_serialized_assignment = PackageRegistryFileBased.__serialize_assignment(assignment)

            sub_client = self.client.dispatch_subdir(dir_name)
            sub_client.mkdir(new_serialized_assignment)
            sub_client.rmdir(serialized_assignment)

    def param_list(self) -> List[str]:
        return self.params

    def param_add(self, name: str, default_value: Optional[str] = "default") -> None:
        if name in self.params:
            raise ValueError("{} already in params".format(name))

        self.client.rmdir(self.__serialize_params(self.params))
        self.params.append(name)
        self.client.mkdir(self.__serialize_params(self.params))

        self.__rename_serialized_assignment(lambda x: x.update({name: default_value}))

    def param_remove(self, name: str) -> None:
        if name not in self.params:
            raise ValueError("{} not in params".format(name))

        new_set_of_serialized_assignment = set()
        for serialized_assignment, dir_name in self.__get_serialized_assignment_to_wrapper_mapping().items():
            assignment = PackageRegistryFileBased.__unserialize_assignment(serialized_assignment)
            del assignment[name]
            new_serialized_assignment = PackageRegistryFileBased.__serialize_assignment(assignment)
            if new_serialized_assignment in new_set_of_serialized_assignment:
                raise NameError("Cannot remove parameter {} since it will cause "
                                "two binary to have the same value".format(name))
            else:
                new_set_of_serialized_assignment.add(new_serialized_assignment)

        self.client.rmdir(self.__serialize_params(self.params))
        self.params.remove(name)
        self.client.mkdir(self.__serialize_params(self.params))

        self.__rename_serialized_assignment(lambda x: x.pop(name))

    def reassign_binary(self, old_assignment: Dict[str, str], new_assignment: Dict[str, str]) -> None:
        if set(new_assignment.keys()) != set(self.params):
            raise KeyError("wrong settings key: {} is not {}".format(sorted(new_assignment.keys()),
                                                                     sorted(self.params)))

        serialized_old_assignment = PackageRegistryFileBased.__serialize_assignment(old_assignment)
        serialized_new_assignment = PackageRegistryFileBased.__serialize_assignment(new_assignment)
        mapping = self.__get_serialized_assignment_to_wrapper_mapping()
        if serialized_old_assignment not in mapping:
            raise ValueError("there is no binary that match the assignment")
        if serialized_new_assignment in mapping:
            raise NameError("there already exist binary with same assignment with the new one")

        sub_client = self.client.dispatch_subdir(mapping[serialized_old_assignment])
        sub_client.rmdir(serialized_old_assignment)
        sub_client.mkdir(serialized_new_assignment)



import argparse
import inspect
from typing import Dict, Callable

from pacco import __version__ as client_version


class CommandAbstract:
    def __init__(self, name, output, remote_manager):
        if name:
            name += " "  # see show help below, we don't want double space for empty `name`
        self.name = name
        self.out = output
        self.rm = remote_manager

    def run(self, *args):
        """
        Entry point for executing commands, dispatcher to class methods.
        """
        if not args:
            self.__show_help()
            return
        command = args[0]
        remaining_args = args[1:]
        commands = self.__get_commands()
        if command not in commands:
            if command in ["-h", "--help"]:
                self.__show_help()
                return
            elif command in ["-v", "--version"]:
                self.out.writeln("Pacco version {}".format(client_version))
                return
            self.out.writeln(
                "'pacco {NAME} {COMMAND}' is an invalid command. See 'pacco {NAME} --help'.".format(
                    NAME=self.name,
                    COMMAND=command),
                error=True
            )
            return
        method = commands[command]
        method(*remaining_args)

    def __get_commands(self) -> Dict[str, Callable]:
        result = {}
        for method_name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not method_name.startswith('_') and method_name not in ["run"]:
                result[method_name] = method
        return result

    def __show_help(self):
        commands = self.__get_commands()
        max_len = max(
            (
                len("pacco {NAME}{COMMAND}".format(NAME=self.name, COMMAND=command))
                for command in commands
            )) + 1
        fmt = '  %-{}s'.format(max_len)
        for command in commands:
            appended_name = "pacco {NAME}{COMMAND}".format(NAME=self.name, COMMAND=command)
            print(fmt % appended_name, end="")
            self.out.writeln(CommandAbstract.__format_docstring(commands[command].__doc__))
        self.out.writeln("")
        self.out.writeln("Pacco {NAME}commands. Type 'pacco {NAME}<command> -h' for help".format(NAME=self.name))

    @staticmethod
    def __format_docstring(docstring: str) -> str:
        if not docstring:
            return ''
        docstring_lines = docstring.split('\n')
        data = []
        for line in docstring_lines:
            line = line.strip()
            data.append(line)
        return ' '.join(data)

    def init_parser(self, method_name: str):
        return argparse.ArgumentParser(prog="pacco {NAME}{COMMAND}".format(NAME=self.name, COMMAND=method_name))
